from time import time
from typing import Coroutine
from asyncio import Task
import asyncio

from aiogram.types import Chat

from limiter import Limiter

GROUPS = ['group', 'supergroup', 'channel']

class LimitCaller:
    
    __slots__ = ('main_limiter', 'chats', 'groups')
    
    def __init__(self) -> None:
        
        """A class that controls the speed of sending requests."""
        
        self.main_limiter = Limiter(30)
        self.chats: dict[int, tuple[Limiter | float | Task]] = {}
        self.groups: dict[int, tuple[Limiter | float | Task]] = {}
        
    async def _delete_task(self, chat: int, 
                          storage: dict[int, tuple[Limiter | float | Task]]):
        
        """Task to clear the chat or group storage"""
        
        await asyncio.sleep(60)
        del storage[chat]
        
    async def _call_with_limit(self, chat_id: int, coro: Coroutine, 
                            storage:  dict[int, tuple[Limiter | float | Task]],
                            rate: int | float, burst: int):
        
        """Calls the api method
        
        Params:
            - chat_id: telegram chat id
            - coro: method
            - storage: chat or group storage
            - rate: call rate for limiters
            - burst: burst for limiters
        """
        
        async with self.main_limiter:
            limiter_list = self.chats.get(chat_id)
            if not limiter_list:
                limiter = Limiter(rate, burst)
                task = asyncio.create_task(self._delete_task(chat_id, storage))
                storage[chat_id] = [limiter, time(), task]
                async with limiter:
                    return await coro
            else:
                limiter_list[2].cancel()
                limiter_list[1] = time()
                limiter_list[2] = asyncio.create_task(self._delete_task(chat_id, storage))
                async with limiter_list[0]:
                    return await coro
    
    async def call(self, chat: Chat, coro: Coroutine):
        
        """Call the method"""
        
        chat_id = chat.id
        chat_type = chat.type
        if chat_type in GROUPS:
            return await self._call_with_limit(chat_id, coro, self.groups, 0.3333, 20)
        else:
            return await self._call_with_limit(chat_id, coro, self.groups, 1, 3)


from typing import Optional, TypeVar
from aiogram.client.session.base import BaseSession

from aiogram.methods import TelegramMethod, GetChat
from aiogram import Bot

from .limit_caller import LimitCaller

T = TypeVar("T")

class LimitedBot(Bot):
    
    async def __call__(self, method: TelegramMethod[T], request_timeout: Optional[int] = None):
        
        coro = self.session(self, method, timeout=request_timeout)
        if hasattr(method, 'chat_id') and not isinstance(method, GetChat):
            caller = self.__dict__.get('caller')
            if not caller:
                caller = LimitCaller()
                self.caller = caller
            chat = await self.get_chat(method.chat_id)
            return await caller.call(chat, coro)
        else:
            return await coro

def path_bot():
    
    """Patches the bot, these changes are not reversible"""
    
    Bot.__call__ = LimitedBot.__call__


from setuptools import setup, find_packages

def readme():
  with open('README.md', 'r') as f:
    return f.read()

setup(
  name='limited_aiogram',
  version='0.1.2',
  author='chazovtema',
  author_email = 'chazovtema@mail.ru',
  description='Limit your api calls to avoid  "Flood control exceeded"',
  long_description=readme(),
  long_description_content_type='text/markdown',
  packages=find_packages(),
  install_requires=['requests>=2.25.1', 'aiogram>=3.0.0', 'limiter>=0.3.1'],
  classifiers=[
    'Programming Language :: Python :: 3.10',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent'
  ],
  keywords= ['aiogram', 'limit'],
  project_urls={"github": "https://github.com/chazovtema/limited_aiogram"},
  python_requires='>=3.10',
)